#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "tsync.h"
#include "tsync_example.h"

#define PTR_DEBUG (0)

#define PTR_POLL_DELAY      (500000) // 500 msec

#define PTR_MAX_RETRY_COUNT (3000)

// Macro to byte swap a 32-bit value
#define SWAPL(lword)                                                    \
    (((lword & 0x000000FF) << 24) |                                     \
     ((lword & 0x0000FF00) <<  8) |                                     \
     ((lword & 0x00FF0000) >>  8) |                                     \
     ((lword & 0xFF000000) >> 24))


int main (int argc, char *argv[])
{
    TSYNC_BoardHandle       hnd;
    int                     rc;
    int                     ret     = 0;
    int                     devIdx;
    char                   *devName = DEFAULT_DEV;
    char                    fullDevName[32];
    char                   *pgmname = argv[0];
    TSYNC_ERROR             err     = TSYNC_SUCCESS;
    FILE                   *inFile;
    unsigned int            fileLen;
    unsigned int            i;
#if (PTR_DEBUG)
    unsigned int            maxRetryCount = 0;
    unsigned int            retryCount = 0;
#endif
    unsigned int            totalCount = 0;
    TSYNC_ULImageIdObj      imgId;
    TSYNC_ULImageHeaderObj  hdr;
    TSYNC_UpdateDataObj     data;
    TSYNC_UpdateEndObj      trl;


    /* If invalid number of arguments... */
    if (argc != 4)
    {
        printf(" Usage: PTR_Upgrade <device index> <inst> <image file>\n");
        return (1);
    }

    /* Construct the TSYNC/TSAT device name */
    devIdx = atoi(argv[1]);
    (void)sprintf( fullDevName, "%s%s%d", DEVICE_ROOT, devName, devIdx );

    /* Open image file */
    inFile = fopen(argv[3], "rb");

    /* Error if can't open file */
    if (inFile == NULL)
    {
        printf("Could not open image file <%s>\n", argv[3]);
        return (1);
    }

    /* Read image header from file */
    (void)fread(&hdr, 1, sizeof(TSYNC_ULImageHeaderObj), inFile);
    hdr.mark = SWAPL(hdr.mark);
    hdr.type = (UL_IMG)SWAPL(hdr.type);
    hdr.len  = SWAPL(hdr.len);

    /* Build image ID */
    imgId.type = (UL_IMG)(hdr.type & 0xFFFF);
    imgId.slot = atoi(argv[2]);

    /* Determine file length */
    (void)fseek(inFile, 0, SEEK_END);
    fileLen = ftell(inFile);

    /* Read image trailer from file */
    (void)fseek(inFile, 0 - (sizeof(TSYNC_UpdateEndObj) - sizeof(UL_IMG)),
                SEEK_CUR);
    (void)fread(&(trl.ver[0]), 1, sizeof(TSYNC_UpdateEndObj) - sizeof(UL_IMG),
                inFile);
    trl.crc = SWAPL(trl.crc);

    /* Print image info */
    printf("\n Image Info for Slot %u:\n", imgId.slot);
    printf("  Mark:    0x%08X\n", (hdr.mark));
    printf("  Type:    0x%08X\n", ((hdr.type) & 0xFFFF));
//    printf("  OC ID:   0x%08X\n", (((hdr.type) >> 16) & 0xFFFF));
    printf("  Length:  0x%08X\n", (hdr.len));
    printf("  Version: %c%c%c%c\n", trl.ver[0], trl.ver[1], trl.ver[2],
           trl.ver[3]);
    printf("  CRC:     0x%08X\n\n", (trl.crc));

    /*
     * open the device
     */
    if ( (rc = TSYNC_open(&hnd, fullDevName)) != TSYNC_SUCCESS )
    {
        /*
         * if errors, close it
         */
        printf("%s: !Could not open <%s> : rc <%d>\n",
                        pgmname, fullDevName, rc);
        fclose(inFile);
        return (1);
    }

    // Send upgrade start transaction
    err = TSYNC_PTR_start(hnd, &imgId, &hdr);

    if (err != TSYNC_SUCCESS)
    {
        printf("  Error: %s.\n", tsync_strerror(err));
        TSYNC_close(hnd);
        fclose(inFile);
        return (1);
    }

    /* Move back to beginning of image */
    (void)fseek(inFile, 0, SEEK_SET);

    for (i = 0; i < fileLen; i += TSYNC_DATA_BLOCK_SIZE)
    {
        /* Build upgrade data message */
        data.type = imgId.type;
        (void)fread(&data.data, 1, TSYNC_DATA_BLOCK_SIZE, inFile);

        err = RC_BUSY;
#if (PTR_DEBUG)
        retryCount=0;
#endif

        while (err == RC_BUSY)
        {
            // Send upgrade data transaction
            // Keep sending the same thing if PTR is busy
            err = TSYNC_PTR_data(hnd, &data);

            if (err == RC_BUSY)
            {
                // Timing System is busy, wait to let low priorty
                // Upgrade task KTS consume previous buffer sent
                usleep(PTR_POLL_DELAY);

#if (PTR_DEBUG)
                retryCount++;
#endif
                totalCount++;
            }

            // Based on above 0.5 sec delay we will timeout after
            // 25 minutes which is 3x the typical ~8 min 31 sec
            // update time. Note if heavy traffic is in ethernet
            // port of 1204-12 PTP card the time may vary and
            // we suggest that if they can't update, to remove
            // ethernet cable or disable slave mode and try again.
            if (totalCount > PTR_MAX_RETRY_COUNT) 
            {
                err = RC_BUSY;
                break;
            }
        }

#if (PTR_DEBUG)
        maxRetryCount = (maxRetryCount < retryCount) ? retryCount : maxRetryCount;

        printf("Loop: Bytes %ld Retry %ld Total %ld\n", i, retryCount, totalCount);
#endif

        if (err != TSYNC_SUCCESS)
        {
            printf("  Error: %s Bytes Sent = %ld Total Retry Count = %ld.\n", 
                   tsync_strerror(err), i, totalCount);
            TSYNC_close(hnd);
            fclose(inFile);
            return (1);
        }

    }
#if (PTR_DEBUG)
    printf("End: Bytes %ld Max Retry %ld Total %ld\n", i, maxRetryCount, totalCount);
#endif
    printf("\n");
    /* Build upgrade end message */
    trl.type = imgId.type;

    // Send upgrade end transaction
    err = TSYNC_PTR_end(hnd, &trl);

    if (err != TSYNC_SUCCESS)
    {
        printf("  Error: %s.\n", tsync_strerror(err));
        TSYNC_close(hnd);
        fclose(inFile);
        return (1);
    }

    if ( (rc = TSYNC_close(hnd)) != TSYNC_SUCCESS )
    {
        (void) printf("%s: error closing <%s> rc: <%d>\n",
                pgmname, fullDevName, rc );
        ret = 1;
    }

    fclose(inFile);

    return (ret);

} /* End - main() */
